@@ -20,6 +20,10 @@ all() ->
      flush_small_post,
      flush_large_post,
      flush_chunked_post,
+     post_multi_different_content_length,
+     post_multi_same_content_length,
+     post_bogus_transfer_encoding,
+     post_bad_order_transfer_encoding,
      flush_small_get,
      flush_large_get,
      flush_chunked_get,
@@ -27,6 +31,7 @@ all() ->
      expires,
      reentrant,
      cgi_redirect,
+     cgi_sanitize,
      php_handler,
      arg_rewrite_rewrite,
      arg_rewrite_redirect,
@@ -49,6 +54,7 @@ all() ->
      accept_ranges,
      status_and_content_length,
      encoded_url,
+     encoded_url_no_docroot_escape,
      header_order,
      extra_response_headers
     ].
@@ -321,6 +327,71 @@ flush_chunked_post(Config) ->
     ?assertEqual(ok, gen_tcp:close(Sock)),
     ok.
 
+post_multi_different_content_length(Config) ->
+    Body = "body doesn't matter for this test",
+    Sz = integer_to_list(length(Body)),
+
+    Port  = testsuite:get_yaws_port(1, Config),
+    Url   = testsuite:make_url(http, "127.0.0.1", Port, "/posttest/"++Sz),
+    BadCL = "42",
+    ClHdr1 = {"Content-Length", Sz},
+    ClHdr2 = {"Content-Length", BadCL},
+    CT     = "binary/octet-stream",
+    ?assertMatch({ok, {{_,400,_}, _, _}}, testsuite:http_post(Url, [ClHdr1,ClHdr2], {CT, Body})),
+    ok.
+
+post_multi_same_content_length(Config) ->
+    File = filename:join(?tempdir(?MODULE), "www/1000.txt"),
+    {ok, FI}  = file:read_file_info(File),
+    Sz = integer_to_list(FI#file_info.size),
+
+    Port  = testsuite:get_yaws_port(1, Config),
+    Url   = testsuite:make_url(http, "127.0.0.1", Port, "/posttest/"++Sz),
+    ClHdr = {"Content-Length", Sz},
+    CT    = "binary/octet-stream",
+    Body  = {fun testsuite:post_file/1, {File,1024000}},
+    ?assertMatch({ok, {{_,200,_}, _, _}}, testsuite:http_post(Url, [ClHdr,ClHdr], {CT, Body})),
+    ok.
+
+post_transfer_encoding(Config, TE, StatusCode) ->
+    Port = testsuite:get_yaws_port(1, Config),
+    Url  = testsuite:make_url(http, "127.0.0.1", Port, "/posttest/chunked/5"),
+    {ok, {Scheme, _, Host, Port, Path, _}} = yaws_dynopts:http_uri_parse(Url),
+    case testsuite:sock_connect(Scheme, Host, Port, [binary,{active,false}], infinity) of
+        {ok, Sock} ->
+            try
+                Req = ["POST " ++ Path ++ " HTTP/1.1\r\n",
+                       "Host: " ++ Host ++ "\r\n",
+                       "User-Agent: Yaws HTTP client\r\n",
+                       "Transfer-Encoding: " ++ TE ++ "\r\n",
+                       "\r\n",
+                       "5\r\n",
+                       "hello\r\n",
+                       "0\r\n"],
+                ok = gen_tcp:send(Sock, Req),
+                ?assertMatch({ok,{{_,StatusCode,_},_,_}}, testsuite:receive_http_response(Sock))
+            after
+                testsuite:sock_close(Sock)
+            end;
+        Error ->
+            ?assertMatch(ok, Error)
+    end,
+    ok.
+
+post_bogus_transfer_encoding(Config) ->
+    %% RFC 7230 secrion 3.3.1 "A server that receives a request
+    %% message with a transfer coding it does not understand SHOULD
+    %% respond with 501 (Not Implemented)."
+    post_transfer_encoding(Config, "bogus", 501).
+
+post_bad_order_transfer_encoding(Config) ->
+    %% RFC 7230 3.3.3 list item 3 states: "If a Transfer-Encoding
+    %% header field is present in a request and the chunked transfer
+    %% coding is not the final encoding, the message body length
+    %% cannot be determined reliably; the server MUST respond with the
+    %% 400 (Bad Request) status code and then close the connection."
+    post_transfer_encoding(Config, "chunked, gzip", 400).
+
 flush_small_get(Config) ->
     File = filename:join(?tempdir(?MODULE), "www/1000.txt"),
     {ok, FI} = file:read_file_info(File),
@@ -497,6 +568,30 @@ cgi_redirect(Config) ->
     ?assert(proplists:is_defined("location", Hdrs)),
     ok.
 
+%% This test attempts to launch a shell command via CGI that writes a
+%% file into the main_SUITE_data/www directory. First verify the file
+%% doesn't exist, then verify that the attempt to execute the command
+%% returns 403, then verify again that the file doesn't exist.
+cgi_sanitize(Config) ->
+    Port = testsuite:get_yaws_port(1, Config),
+    File = "you-got-hacked",
+    Filename = filename:join([?data_srcdir(?MODULE), "www", File]),
+    _ = file:delete(Filename),
+    ?assertEqual({error, enoent}, file:read_file_info(Filename)),
+    [_|FileSegments] = filename:split(Filename),
+    FileRejoined = yaws:join_sep(FileSegments, "$Z''"),
+    Path = lists:flatten(["/cgi-bin/%22%60Z=$(pwd%7Ccut%20-c1);echo%20ouch%3E$Z''",
+                          FileRejoined,
+                          "%60%22"]),
+    Url1 = testsuite:make_url(http, "127.0.0.1", Port, "/"++File),
+    Url2 = testsuite:make_url(http, "127.0.0.1", Port, Path),
+
+    ?assertMatch({ok, {{_,404,_}, _, _}}, testsuite:http_get(Url1)),
+    ?assertMatch({ok, {{_,403,_}, _, _}}, testsuite:http_get(Url2)),
+    ?assertMatch({ok, {{_,404,_}, _, _}}, testsuite:http_get(Url1)),
+    ?assertEqual({error, enoent}, file:read_file_info(Filename)),
+    ok.
+
 php_handler(Config) ->
     Port = testsuite:get_yaws_port(1, Config),
     Url = testsuite:make_url(http, "127.0.0.1", Port, "/test.php"),
@@ -807,6 +902,28 @@ encoded_url(Config) ->
     ?assertEqual(PathInfo, proplists:get_value("x-pathinfo", Hdrs)),
     ok.
 
+encoded_url_no_docroot_escape(Config) ->
+    Port = testsuite:get_yaws_port(1, Config),
+    Url1 = testsuite:make_url(http, "127.0.0.1", Port, "/..%2Fyaws.conf"),
+    {ok, {{_,404,_}, _, _}} = testsuite:http_get(Url1),
+    Url2 = testsuite:make_url(http, "127.0.0.1", Port, "/..%252Fyaws.conf"),
+    {ok, {{_,404,_}, _, _}} = testsuite:http_get(Url2),
+    Url3 = testsuite:make_url(http, "127.0.0.1", Port, "/%2E%2E/yaws.conf"),
+    {ok, {{_,404,_}, _, _}} = testsuite:http_get(Url3),
+    Url4 = testsuite:make_url(http, "127.0.0.1", Port, "/%252E%252E/yaws.conf"),
+    {ok, {{_,404,_}, _, _}} = testsuite:http_get(Url4),
+    %% the next two tests verify only one level of URL decoding occurs
+    Url5 = testsuite:make_url(http, "127.0.0.1", Port, "/%252fexisting-file"),
+    {ok, {{_,200,_}, _, _}} = testsuite:http_get(Url5),
+    Url6 = testsuite:make_url(http, "127.0.0.1", Port, "/%25252fexisting-file"),
+    {ok, {{_,200,_}, _, _}} = testsuite:http_get(Url6),
+    %% path traversal with prepended directory
+    Url7 = testsuite:make_url(http, "127.0.0.1", Port, "/dir/../%25252fexisting-file"),
+    {ok, {{_,200,_}, _, _}} = testsuite:http_get(Url7),
+    Url8 = testsuite:make_url(http, "127.0.0.1", Port, "/dir/..%2F../yaws.conf"),
+    {ok, {{_,404,_}, _, _}} = testsuite:http_get(Url8),
+    ok.
+
 %% RFC 7230 section 3.3.2 discusses header order, and its opening
 %% statement says that header order is not significant. But it then
 %% says that it's good practice to order general headers and control
@@ -865,6 +982,8 @@ prepare_docroots() ->
     ok = file:write_file(filename:join(WWW, "3000.txt"),  lists:duplicate(3000,  Data), [write]),
     ok = file:write_file(filename:join(WWW, "10000.txt"), lists:duplicate(10000, Data), [write]),
 
+    ok = file:write_file(filename:join(WWW, "%252fexisting-file"), "existing file", [write]),
+    ok = file:write_file(filename:join(WWW, "%2fexisting-file"), "existing file", [write]),
     ok.
 
 allow_connects([], _) ->
